import os
import numpy as np
import iris
import iris.cube
import iris.analysis
import warnings
import numba
import datetime
from numba import jit, njit
from numba import vectorize
from numba import guvectorize
from numba import cuda
from numba import int32, int64, float32, float64
from numba.cuda.cudadrv import driver

MESH_FILE = "/esarchive/autosubmit/con_files/mesh_mask_nemo.Ec3.2_O1L75.nc"
TOS_FILE = "/esarchive/exp/ecearth/a13c/cmorfiles/CMIP/EC-Earth-Consortium"\
            "/EC-Earth3-LR/historical/r1i1p1f1/Omon/tos/gn/v20180501"\
            "/tos_Omon_EC-Earth3-LR_historical_r1i1p1f1_gn_195001-195012.nc"
THETAO_FILE = "/esarchive/exp/ecearth/a13c/cmorfiles/CMIP/EC-Earth-Consortium"\
              "/EC-Earth3-LR/historical/r1i1p1f1/Omon/thetao/gn/v20180501/"\
              "thetao_Omon_EC-Earth3-LR_historical_r1i1p1f1_gn_"\
              "195001-195012.nc"
REGIONS_FILE = "/esarchive/autosubmit/con_files/mask.regions.Ec3.2_O1L75.nc"


def main():
    e1t, e2t, e3t_0, mask = load_mesh()
    weights = compute_weights_2D(mask, e1t, e2t)
    weights_3D = compute_weights_3D(mask, e1t, e2t, e3t_0)
    del mask, e1t, e2t, e3t_0
    tos = load_tos()
    for i in range(10):
        start_cpu = datetime.datetime.now()
        regmean_tos_cpu = compute_regmean_2D_cpu(tos, weights)
        end_cpu = datetime.datetime.now() - start_cpu
        start_gpu = datetime.datetime.now()
        regmean_tos_gpu = compute_regmean_2D_gpu(tos, weights)
        end_gpu = datetime.datetime.now() - start_gpu
        print('Computing time 2D var CPU: {0} s'.format(end_cpu))
        print('Computing time 2D var GPU: {0} s'.format(end_gpu))

    del tos
    start_cpu_3D = datetime.datetime.now()
    thetao = load_thetao()
    for i in range(10):
        regmean_thetao_cpu = compute_regmean_3D_cpu(thetao, weights_3D)
        end_cpu_3D = datetime.datetime.now() - start_cpu_3D
        start_gpu_3D = datetime.datetime.now()
        regmean_thetao_gpu = compute_regmean_gpu_3D(thetao, weights_3D)
        end_gpu_3D = datetime.datetime.now() - start_gpu_3D
        print('Computing time 3D var CPU: {0} s'.format(end_cpu_3D))
        print('Computing time 3D var CPU: {0} s'.format(end_gpu_3D))


def load_mesh():
    e1t = iris.load_cube(MESH_FILE, 'e1t').data.astype(np.float32)
    e2t = iris.load_cube(MESH_FILE, 'e2t').data.astype(np.float32)
    e3t = iris.load_cube(MESH_FILE, 'e3t_0').data.astype(np.float32)
    mask = iris.load_cube(MESH_FILE, 'tmask')[0, 0, :].data.astype(np.float32)
    return e1t, e2t, e3t, mask


def load_masks():
    basins = {}
    cubes = iris.load(REGIONS_FILE)
    for cube in cubes:
        name = cube.name()
        if name in ('nav_lat', 'nav_lon'):
            continue
        cube = cube.extract(iris.Constraint(z=1))
        basins[name] = cube.data.astype(np.float32)
    return basins


def load_tos():
    tos = iris.load_cube(TOS_FILE, 'sea_surface_temperature')
    tos_data = tos.data.astype(np.float32)
    return tos_data


def load_thetao():
    thetao = iris.load_cube(THETAO_FILE, 'sea_water_potential_temperature')
    thetao_data = np.ma.filled(thetao.data, 0.0).astype(np.float32)
    return thetao_data


def compute_regmean_2D_cpu(var, weights):
    weighted_var, vector_weight = variable_weight_2D(var, weights)
    reduced_weight = sum_red.reduce(vector_weight, axis=0)
    reduced_weighted_var = sum_red.reduce(weighted_var, axis=1)
    regmean = regmean_cpu(reduced_weighted_var, reduced_weight)
    return regmean


def compute_regmean_3D_cpu(var, weights):
    weighted_var, vector_weight = variable_weight_3D(var, weights)
    reduced_weight = sum_red.reduce(vector_weight, axis=0)
    reduced_weighted_var = sum_red.reduce(weighted_var, axis=1)
    regmean = regmean_cpu(reduced_weighted_var, reduced_weight)
    return regmean


def compute_regmean_2D_gpu(var, weights):
    times = var.shape[0]
    lats = var.shape[1]
    lons = var.shape[2]

    block = (128, 1, 1)
    grid_size = (lons // block[0]) + 1
    grid = (grid_size, lats, times)

    gpu_var = cuda.to_device(var.astype(np.float32))
    gpu_weights = cuda.to_device(weights.astype(np.float32))
    gpu_weighted_var = cuda.device_array((times, lats*lons), dtype=np.float32)
    gpu_vector_weight = cuda.device_array(lats*lons, dtype=np.float32)

    variable_weight_cuda[grid, block](gpu_var, gpu_weights, gpu_weighted_var,
                                      gpu_vector_weight)
    del gpu_var, gpu_weights

    reduced_weight = sum_red_cuda(gpu_vector_weight)
    del gpu_vector_weight

    regmean = np.empty(times, dtype=np.float32)
    for time in range(times):
        reduced_weighted_var = sum_red_cuda(gpu_weighted_var[time, :])
        regmean[time] = reduced_weighted_var / reduced_weight
    del gpu_weighted_var

    return regmean


def compute_regmean_gpu_3D(var, weights):
    times = var.shape[0]
    levs = var.shape[1]
    lats = var.shape[2]
    lons = var.shape[3]

    block = (128, 1, 1)
    grid_size = (lons // block[0]) + 1
    grid = (grid_size, lats, levs)

    gpu_var = cuda.to_device(var.astype(np.float32))
    gpu_weights = cuda.to_device(weights.astype(np.float32))

    gpu_weighted_var = cuda.device_array((times, levs*lats*lons),
                                         dtype=np.float32)
    gpu_vector_weight = cuda.device_array(levs*lats*lons, dtype=np.float32)

    variable_weight_cuda_3D[grid, block](gpu_var, gpu_weights,
                                         gpu_weighted_var, gpu_vector_weight)
    del gpu_var, gpu_weights

    reduced_weight = sum_red_cuda(gpu_vector_weight)
    del gpu_vector_weight

    regmean = np.empty(times, dtype=np.float32)
    for time in range(times):
        reduced_weighted_var = sum_red_cuda(gpu_weighted_var[time, :])
        regmean[time] = reduced_weighted_var / reduced_weight
    del gpu_weighted_var

    return regmean


@vectorize(['float32(float32, float32, float32)'], target='cpu')
def compute_weights_2D(mask, e1t, e2t):
    weights = mask*e1t*e2t
    return weights


@vectorize(['float32(float32, float32, float32, float32)'], target='cpu')
def compute_weights_3D(mask, e1t, e2t, e3t):
    weights = mask*e1t*e2t*e3t
    return weights


@njit
def variable_weight_2D(var, weight):
    times = var.shape[0]
    lats = var.shape[1]
    lons = var.shape[2]
    weighted_var = np.empty((times, lats*lons))
    vector_weight = np.empty(lats*lons)
    for t in range(times):
        for j in range(lats):
            for i in range(lons):
                x = i+j*var.shape[2]
                weight_value = weight[0, j, i]
                value = var[t, j, i] * weight_value
                vector_weight[x] = weight_value
                weighted_var[t, x] = value
    return weighted_var, vector_weight


@njit
def variable_weight_3D(var, weight):
    times = var.shape[0]
    levs = var.shape[1]
    lats = var.shape[2]
    lons = var.shape[3]
    weighted_var = np.empty((times, levs*lats*lons))
    vector_weight = np.empty(levs*lats*lons)
    for t in range(times):
        for l in range(levs):
            for j in range(lats):
                for i in range(lons):
                    x = i+lons*(j+lats*l)
                    weight_value = weight[0, l, j, i]
                    value = var[t, l, j, i] * weight_value
                    vector_weight[x] = weight_value
                    weighted_var[t, x] = value
    return weighted_var, vector_weight


@cuda.jit
def variable_weight_cuda(var, weight, weighted_var, vector_weight):
    i, j, t = cuda.grid(3)
    x = i+j*var.shape[2]
    if(i >= var.shape[2]):
        return
    vector_weight[x] = 0.0
    weighted_var[t, x] = 0.0
    vector_weight[x] = weight[0, j, i]
    weighted_var[t, x] = var[t, j, i] * weight[0, j, i]


@cuda.jit
def variable_weight_cuda_3D(var, weight, weighted_var, vector_weight):
    i, j, lev = cuda.grid(3)
    x = i+var.shape[3]*(j+var.shape[2]*lev)
    if(i >= var.shape[3]):
        return
    vector_weight[x] = 0.0
    vector_weight[x] = weight[0, lev, j, i]
    for t in var.shape[0]:
        weighted_var[t, x] = 0.0
        weighted_var[t, x] = var[t, lev, j, i] * weight[0, lev, j, i]


@vectorize(['float64(float32, float32)', 'float64(float64, float64)'],
           target='cpu')
def sum_red(x, y):
    return x+y


@vectorize(['float32(float64, float64)'], target='cpu')
def regmean_cpu(weighted_var, weights):
    regmean = weighted_var/weights
    return regmean


@vectorize(['float32(float64, float64)'], target='cuda')
def regmean_cuda(weighted_var, weights):
    regmean = weighted_var/weights
    return regmean


@cuda.reduce
def sum_red_cuda(x, y):
    return x+y


if __name__ == '__main__':
    main()
